package fr.emanation.ij.beanshellbox;

import bsh.EvalError;
import bsh.Interpreter;
import bsh.util.JConsole;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.openapi.roots.OrderRootType;
import com.intellij.openapi.util.IconLoader;
import com.intellij.openapi.vfs.JarFileSystem;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VirtualFileSystem;
import com.intellij.openapi.wm.ToolWindow;
import com.intellij.openapi.wm.ToolWindowAnchor;
import com.intellij.openapi.wm.ToolWindowManager;
import com.intellij.openapi.wm.WindowManager;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import java.util.zip.ZipFile;
import java.net.MalformedURLException;

/**
 * <dl>
 * <dt><b>Creation date :</b></dt>
 * <dd>20 sept. 2006</dd>
 * This plugin creates a new tool window which integrates the standard BeanShell console.
 * BeanShell console allows interpreting beanshell scripts as well as plain Java code.
 * Thus this plugin is useful for anyone wanting a scratchpad for testing bunches of code.
 * For conveniences, the plugin allows reseting the execution environment and loading scripts from
 * files.
 *
 * @author Cedric CHAMPEAU &lt;cedric.champeau@laposte.net&gt;
 * @lastModified September 20th, 2006
 * </dl>
 */
public class BeanShellConsole implements ProjectComponent {
	public static final String TOOLWINDOW_ID = "BeanShell Box";

	private Project theProject;
	private JConsole theConsole;
	private Interpreter theInterpreter;

	public BeanShellConsole(Project aProject) {
		theProject = aProject;
	}

	// --------------------- ProjectComponent implementation ---------------------

	public void initComponent() {
		theConsole = new JConsole();
	}

	public void disposeComponent() {
		theConsole.removeAll();
		theConsole = null;
	}

	public String getComponentName() {
		return "BeanShellConsole";
	}

	public void projectOpened() {
		initUI();
		resetInterpreter();
	}

	public void projectClosed() {
		ToolWindowManager manager = ToolWindowManager.getInstance(theProject);
		manager.unregisterToolWindow(TOOLWINDOW_ID);
	}

	// ----------------- UI related methods --------------------------------

	/**
	 * Builds the console UI
	 */
	private void initUI() {
		JPanel panel = new JPanel(new BorderLayout());
		panel.add(theConsole, BorderLayout.CENTER);
		JToolBar toolbar = new JToolBar();
		toolbar.setOrientation(JToolBar.VERTICAL);
		toolbar.setFloatable(false);
		toolbar.add(new ResetConsoleAction());
		toolbar.add(new OpenScriptAction());
		panel.add(toolbar, BorderLayout.WEST);
		ToolWindowManager manager = ToolWindowManager.getInstance(theProject);
		ToolWindow toolwindow = manager.registerToolWindow(TOOLWINDOW_ID, panel, ToolWindowAnchor.BOTTOM);
		toolwindow.setTitle("Console");
		toolwindow.setIcon(IconLoader.getIcon("/fileTypes/text.png"));
	}

	public void expandPanel() {
		ToolWindowManager manager = ToolWindowManager.getInstance(theProject);

		ToolWindow tw = manager.getToolWindow(TOOLWINDOW_ID);
		if (!tw.isVisible()) {
			tw.show(null);
		}
		theConsole.requestFocus();
	}

	// ----------------- interpreter related methods -----------------------

	/**
	 * Clears all variables, methods and imports in the current console scope.
	 * Creates the interpreter if it does not exist.
	 */
	private void resetInterpreter() {
		if (theInterpreter == null) {
			theInterpreter = new Interpreter(theConsole);
			
			new Thread(theInterpreter).start();
		} else {
			try {
				theInterpreter.eval("clear();");
				logError("Cleared all variables, methods and imports.");
				emulateEnterPressed();
			} catch (EvalError aEvalError) {
				logError(aEvalError.getScriptStackTrace());
			}
		}
		importClassPathes();
	}

	/**
	 * Makes the project classpathes available to BeanShell
	 */
	private void importClassPathes() {
		final ModuleManager manager = ModuleManager.getInstance(theProject);
		final Module[] modules = manager.getModules();
		for (Module module : modules) {
			ModuleRootManager mrm = ModuleRootManager.getInstance(module);
			final VirtualFile[] files = mrm.getFiles(OrderRootType.CLASSES_AND_OUTPUT);
			for (VirtualFile virtualFile : files) {
				if (!virtualFile.isValid())
					continue;
				final VirtualFileSystem fileSystem = virtualFile.getFileSystem();
				final File file;
				if (fileSystem instanceof LocalFileSystem)
					file = new File(virtualFile.getPath().replace('/', File.separatorChar));
				else if (fileSystem instanceof JarFileSystem) {
					final JarFileSystem jarFileSystem = (JarFileSystem) fileSystem;
					try {
						final ZipFile jarFile = jarFileSystem.getJarFile(virtualFile);
						file = new File(jarFile.getName());
						if (isJDKJar(file)) continue;
					} catch (IOException e) {
						logError("Error getting jar file: " + e);
						continue;
					}
				} else
					continue;
				try {
					theConsole.error("Loading " + file.getAbsolutePath()+" ...");
					theConsole.println();
					theInterpreter.eval("addClassPath(new URL(\""+file.toURL()+"\"));");
				} catch (EvalError aEvalError) {
					logError(aEvalError.getScriptStackTrace());
				} catch (MalformedURLException e) {
					logError(e.getMessage());
				}
			}
		}
	}

	/**
	 * Tests if the specified Jar file is a JRE one. Only based on JAR path name !
	 * todo : improve detection
	 * @param aFile Jar file to test
	 * @return <code>boolean</code> true if the Jar file is suspected to be a JRE file
	 */
	private boolean isJDKJar(File aFile) {
		if (aFile==null) return false;
		final String pattern = "jre"+File.separator+"lib";
		String sn = aFile.getAbsolutePath().toLowerCase();
		return sn.contains(pattern);
	}
	/**
	 * Allows passing script to the interpreter.
	 *
	 * @param aScript the script to be interpreted
	 */
	public void eval(String aScript) {
		theConsole.println(aScript);
		emulateEnterPressed();
		try {
			theInterpreter.eval(aScript);
		} catch (EvalError aEvalError) {
			logError(aEvalError.getScriptStackTrace());
		}
	}

	/**
	 * This method sends a "VK_ENTER" event to the console in order to make it refresh its display
	 * todo : is there any better way of refreshing the display ?
	 */
	private void emulateEnterPressed() {
		long now = System.currentTimeMillis();
		theConsole.keyPressed(new KeyEvent(theConsole, KeyEvent.KEY_PRESSED, now, 0, KeyEvent.VK_ENTER, '\n', KeyEvent.KEY_LOCATION_STANDARD));
		theConsole.keyReleased(new KeyEvent(theConsole, KeyEvent.KEY_RELEASED, now, 0, KeyEvent.VK_ENTER, '\n', KeyEvent.KEY_LOCATION_STANDARD));
		theConsole.keyTyped(new KeyEvent(theConsole, KeyEvent.KEY_TYPED, now, 0, KeyEvent.VK_UNDEFINED, '\n', KeyEvent.KEY_LOCATION_UNKNOWN));
	}

	/**
	 * Sends an error to the console
	 *
	 * @param anError the error text
	 */
	private void logError(String anError) {
		theConsole.error(anError);
		emulateEnterPressed();
	}

	// --------------- UI Actions -----------------------------------------

	/**
	 * Allows reseting the beanshell environment
	 */
	private class ResetConsoleAction extends AbstractAction {

		public ResetConsoleAction() {
			super("Reset", IconLoader.getIcon("/actions/sync.png"));
			putValue(AbstractAction.SHORT_DESCRIPTION, "Reset environment");
		}

		public void actionPerformed(ActionEvent aActionEvent) {
			resetInterpreter();
			theConsole.requestFocus();
		}
	}

	/**
	 * Allows opening an executing a beanshell script
	 */
	private class OpenScriptAction extends AbstractAction {

		public OpenScriptAction() {
			super("Open", IconLoader.getIcon("/actions/menu-open.png"));
			putValue(AbstractAction.SHORT_DESCRIPTION, "Open BeanShell script");
		}


		public void actionPerformed(ActionEvent aActionEvent) {
			JFileChooser fc = new JFileChooser();
			fc.setMultiSelectionEnabled(false);
			fc.setFileFilter(new FileFilter() {
				public boolean accept(File aFile) {
					return aFile.isDirectory() || aFile.getName().toLowerCase().endsWith(".bsh");
				}

				public String getDescription() {
					return "BeanShell scripts (.bsh)";
				}
			});
			int result = fc.showOpenDialog(WindowManager.getInstance().suggestParentWindow(theProject));
			if (result == JFileChooser.APPROVE_OPTION) {
				File script = fc.getSelectedFile();
				try {
					theInterpreter.source(script.getAbsolutePath());
				} catch (IOException e) {
					logError(e.getMessage());
				} catch (EvalError aEvalError) {
					logError(aEvalError.getScriptStackTrace());
				}
			}
		}
	}
}
